/*
 * Generacion de codigo.
 */
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "asm.h"
#include "parse.h"

#define TFR(_)		_

struct linea_s {
	struct linea_s *siguiente;
	char linea[0];
} *principio, *fin;

/*
 * Buffer para lineas de codigo ensamblador.
 */
char buffer[512];

/*
 * Ultimo numero de linea en fichero fuente.
 */
int l_lin = 1;

/*
 * A base de este numero se generan etiquetas no repetidas
 * en el codigo. Se incrementa despues de cada uso.
 */
int etiqueta = 0;

void asm_inicio() {
	principio = fin = malloc(
			sizeof(struct linea_s) + strlen(CABECERA) + 1);
	fin->siguiente = 0;
	strcpy(fin->linea, CABECERA);
}

void asm_linea() {
	char lin[128];
	l_lin = linea;
	sprintf(lin, "%%line %i+0 %s", l_lin, input_fn);
	fin = fin->siguiente = malloc(
			sizeof(struct linea_s) + strlen(lin) + 1);
	fin->siguiente = 0;
	strcpy(fin->linea, lin);
}

void asm_anadir() {
	if (l_lin != linea)
		asm_linea();
	fin = fin->siguiente = malloc(
			sizeof(struct linea_s) + strlen(buffer) + 1);
	fin->siguiente = 0;
	strcpy(fin->linea, buffer);
}

void asm_guardar(int fd) {
	struct linea_s *lin = principio;
	while (lin) {
		TFR(write(fd, lin->linea, strlen(lin->linea)));
		TFR(write(fd, "\n", 1));
		lin = lin->siguiente;
	}
}

void asm_fin() {
	while (fin = principio) {
		principio = fin->siguiente;
		free(fin);
	}
}

void asm_segmento_datos() {
	sprintf(buffer, "segment .data");
	asm_anadir();
}

void asm_segmento_codigo() {
	sprintf(buffer, "segment .text");
	asm_anadir();
	sprintf(buffer, "\tglobal main");
	asm_anadir();
	sprintf(buffer, "\textern " EXTERN0);
	asm_anadir();
	sprintf(buffer, "\textern " EXTERN1);
	asm_anadir();
}

void asm_segmento_fin() {
	sprintf(buffer, "\tret");
	asm_anadir();
}

void asm_main() {
	sprintf(buffer, "main:");
	asm_anadir();
}

void asm_variable(token_t tipo) {
	int i, refs;
	/* Bucle para punteros */
	for (refs = tipo.num_ref; refs >= 0; refs --) {
		strcpy(buffer, "\t");

		for (i = 0; i <= refs; i ++)
			strcat(buffer, VAR_PREFIJO);
		strcat(buffer, tipo.clave);
		strcat(buffer, " dd ");
		if (refs < tipo.num_ref) {
			for (i = refs + 2; i; i --)
				strcat(buffer, VAR_PREFIJO);
			strcat(buffer, tipo.clave);
		} else
			strcat(buffer, "0");

		asm_anadir();
	}
}

void asm_variable_push(const char *nombre) {
	sprintf(buffer, "\tpush dword " VAR_PREFIJO "%s", nombre);
	asm_anadir();
}

void asm_variable_funcion(token_t tipo) {
	if (tipo.clase == variable)
		sprintf(buffer, "\tlea  eax, [ebp - %i]",
				(tipo.num_param + 1) * 4);
	else
		sprintf(buffer, "\tlea  eax, [ebp + %i]",
				(tipo.num_param + 1) * 4);
	asm_anadir();
	sprintf(buffer, "\tpush eax");
	asm_anadir();
}

void asm_desreferenciacion() {
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	sprintf(buffer, "\tpush dword [eax]");
	asm_anadir();
}

void asm_asignacion() {
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	sprintf(buffer, "\tpop  edx");
	asm_anadir();
	sprintf(buffer, "\tmov  [edx], eax");
	asm_anadir();
}

void asm_input(token_t tipo) {
	/*
	 * No hace falta sacar nada de la pila porque la
	 * direccion de la variable debe ya estar en la pila.
	 */
	if (tipo.tipo == t_int)
		sprintf(buffer, "\tcall lee_entero");
	else
		sprintf(buffer, "\tcall lee_booleano");
	asm_anadir();
	sprintf(buffer, "\tadd  esp, 4");
	asm_anadir();
}

void asm_output(token_t tipo) {
	/*
	 * No hace falta sacar nada de la pila porque el
	 * valor de la expresion debe ya estar en la pila.
	 */
	if (tipo.tipo == t_int)
		sprintf(buffer, "\tcall imprime_entero");
	else
		sprintf(buffer, "\tcall imprime_booleano");
	asm_anadir();
	sprintf(buffer, "\tadd  esp, 4");
	asm_anadir();
	sprintf(buffer, "\tcall imprime_fin_linea");
	asm_anadir();
}

void asm_valor_push(int valor) {
	sprintf(buffer, "\tpush dword %i", valor);
	asm_anadir();
}

void asm_suma(token_t tipo) {
	sprintf(buffer, "\tpop  edx");
	asm_anadir();
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	if (tipo.tipo == t_int)
		sprintf(buffer, "\tadd  eax, edx");
	else
		sprintf(buffer, "\tor   eax, edx");
	asm_anadir();
	sprintf(buffer, "\tpush eax");
	asm_anadir();
}

void asm_resta(token_t tipo) {
	sprintf(buffer, "\tpop  edx");
	asm_anadir();
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	if (tipo.tipo == t_int)
		sprintf(buffer, "\tsub  eax, edx");
	else
		sprintf(buffer, "\txor  eax, edx");
	asm_anadir();
	sprintf(buffer, "\tpush eax");
	asm_anadir();
}

void asm_multiplicacion(token_t tipo) {
	sprintf(buffer, "\tpop  edx");
	asm_anadir();
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	if (tipo.tipo == t_int)
		sprintf(buffer, "\tmul  edx");
	else
		sprintf(buffer, "\tand  eax, edx");
	asm_anadir();
	sprintf(buffer, "\tpush eax");
	asm_anadir();
}

void asm_negacion(token_t tipo) {
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	if (tipo.tipo == t_int)
		sprintf(buffer, "\tneg  eax");
	else
		sprintf(buffer, "\txor  eax, 1");
	asm_anadir();
	sprintf(buffer, "\tpush eax");
	asm_anadir();
}

void asm_igualdad() {
	sprintf(buffer, "\tpop  edx");
	asm_anadir();
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	sprintf(buffer, "\tcmp  eax, edx");
	asm_anadir();
	sprintf(buffer, "\tjne  near l%i", etiqueta);
	asm_anadir();
	sprintf(buffer, "\tpush dword 1");
	asm_anadir();
	sprintf(buffer, "\tjmp  near l%i", etiqueta + 1);
	asm_anadir();
	sprintf(buffer, "l%i:", etiqueta);
	asm_anadir();
	sprintf(buffer, "\tpush dword 0");
	asm_anadir();
	sprintf(buffer, "l%i:", etiqueta + 1);
	asm_anadir();
	etiqueta += 2;
}

void asm_desigualdad() {
	sprintf(buffer, "\tpop  edx");
	asm_anadir();
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	sprintf(buffer, "\tcmp  eax, edx");
	asm_anadir();
	sprintf(buffer, "\tje   near l%i", etiqueta);
	asm_anadir();
	sprintf(buffer, "\tpush dword 1");
	asm_anadir();
	sprintf(buffer, "\tjmp  near l%i", etiqueta + 1);
	asm_anadir();
	sprintf(buffer, "l%i:", etiqueta);
	asm_anadir();
	sprintf(buffer, "\tpush dword 0");
	asm_anadir();
	sprintf(buffer, "l%i:", etiqueta + 1);
	asm_anadir();
	etiqueta += 2;
}

void asm_menor() {
	sprintf(buffer, "\tpop  edx");
	asm_anadir();
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	sprintf(buffer, "\tcmp  eax, edx");
	asm_anadir();
	sprintf(buffer, "\tjnl  near l%i", etiqueta);
	asm_anadir();
	sprintf(buffer, "\tpush dword 1");
	asm_anadir();
	sprintf(buffer, "\tjmp  near l%i", etiqueta + 1);
	asm_anadir();
	sprintf(buffer, "l%i:", etiqueta);
	asm_anadir();
	sprintf(buffer, "\tpush dword 0");
	asm_anadir();
	sprintf(buffer, "l%i:", etiqueta + 1);
	asm_anadir();
	etiqueta += 2;
}

void asm_menor_igual() {
	sprintf(buffer, "\tpop  edx");
	asm_anadir();
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	sprintf(buffer, "\tcmp  eax, edx");
	asm_anadir();
	sprintf(buffer, "\tjnle near l%i", etiqueta);
	asm_anadir();
	sprintf(buffer, "\tpush dword 1");
	asm_anadir();
	sprintf(buffer, "\tjmp  near l%i", etiqueta + 1);
	asm_anadir();
	sprintf(buffer, "l%i:", etiqueta);
	asm_anadir();
	sprintf(buffer, "\tpush dword 0");
	asm_anadir();
	sprintf(buffer, "l%i:", etiqueta + 1);
	asm_anadir();
	etiqueta += 2;
}

void asm_mayor() {
	sprintf(buffer, "\tpop  edx");
	asm_anadir();
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	sprintf(buffer, "\tcmp  eax, edx");
	asm_anadir();
	sprintf(buffer, "\tjng  near l%i", etiqueta);
	asm_anadir();
	sprintf(buffer, "\tpush dword 1");
	asm_anadir();
	sprintf(buffer, "\tjmp  near l%i", etiqueta + 1);
	asm_anadir();
	sprintf(buffer, "l%i:", etiqueta);
	asm_anadir();
	sprintf(buffer, "\tpush dword 0");
	asm_anadir();
	sprintf(buffer, "l%i:", etiqueta + 1);
	asm_anadir();
	etiqueta += 2;
}

void asm_mayor_igual() {
	sprintf(buffer, "\tpop  edx");
	asm_anadir();
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	sprintf(buffer, "\tcmp  eax, edx");
	asm_anadir();
	sprintf(buffer, "\tjnge near l%i", etiqueta);
	asm_anadir();
	sprintf(buffer, "\tpush dword 1");
	asm_anadir();
	sprintf(buffer, "\tjmp  near l%i", etiqueta + 1);
	asm_anadir();
	sprintf(buffer, "l%i:", etiqueta);
	asm_anadir();
	sprintf(buffer, "\tpush dword 0");
	asm_anadir();
	sprintf(buffer, "l%i:", etiqueta + 1);
	asm_anadir();
	etiqueta += 2;
}

void asm_if(int etiqueta) {
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	sprintf(buffer, "\tcmp  eax, 0");
	asm_anadir();
	sprintf(buffer, "\tje   near lif%i", etiqueta);
	asm_anadir();
}

void asm_else(int etiqueta) {
	sprintf(buffer, "\tjmp  near lfi%i", etiqueta);
	asm_anadir();
	sprintf(buffer, "lif%i:", etiqueta);
	asm_anadir();
}

void asm_fi(int etiqueta) {
	sprintf(buffer, "lif%i:", etiqueta);
	asm_anadir();
}

void asm_else_fi(int etiqueta) {
	sprintf(buffer, "lfi%i:", etiqueta);
	asm_anadir();
}

void asm_llamada(token_t func) {
	sprintf(buffer, "\tcall " VAR_PREFIJO "%s", func.clave);
	asm_anadir();
	sprintf(buffer, "\tadd  esp, %i", func.num_param << 2);
	asm_anadir();
	sprintf(buffer, "\tpush eax");
	asm_anadir();
}

void asm_return() {
	sprintf(buffer, "\tpop  eax");
	asm_anadir();
	sprintf(buffer, "\tmov  esp, ebp");
	asm_anadir();
	sprintf(buffer, "\tpop  ebp");
	asm_anadir();
	sprintf(buffer, "\tret");
	asm_anadir();
}

void asm_funcion(token_t tipo, int variables) {
	sprintf(buffer, VAR_PREFIJO "%s:", tipo.clave);
	asm_anadir();
	sprintf(buffer, "\tpush ebp");
	asm_anadir();
	sprintf(buffer, "\tmov  ebp, esp");
	asm_anadir();
	sprintf(buffer, "\tsub  esp, %i", variables * 4);
	asm_anadir();
}

void asm_funcion_end() {
	sprintf(buffer, "\txor  eax, eax");
	asm_anadir();
	sprintf(buffer, "\tmov  esp, ebp");
	asm_anadir();
	sprintf(buffer, "\tpop  ebp");
	asm_anadir();
	sprintf(buffer, "\tret");
	asm_anadir();
}
