%{
	#include <stdlib.h>

	#include "parse.h"
	#include "simbolos.h"
	#include "asm.h"

	#define YYSTYPE token_t
	#define yyerror error_sintactico
	#define yyparse parse

	/*
	 * Variable global que hace posible la herencia de
	 * atributos entre hermanos en la produccion "declaration".
	 */
	token_t last_mode;

	/*
	 * Produce una copia dinamica del valor semantico "t"
	 * lista para insertar en la tabla de simbolos.
	 */
	token_t *copia(token_t t);

	/*
	 * La "profundidad" del contexto actual, 0 para main, 1
	 * para funcion en main, etc
	 */
	int nivel = 0;

	/*
	 * Como solamente puede haber un nivel de funciones
	 * (y ambitos) usamos variables globales para definiciones
	 * de funciones.
	 */
	int num_param;
	token_t retorno;

	/*
	 * Nombre del ambito en el que se ha encontrado el ultimo
	 * identificador leido.
	 */
	char *ambito;
%}

%token T_NUMBER
%token T_ID

%token T_BEGIN
%token T_END
%token T_BOOL
%token T_INT
%token T_REF
%token T_FUNCTION
%token T_RETURN
%token T_IF
%token T_THEN
%token T_FI
%token T_ELSE
%token T_WHILE
%token T_DO
%token T_INPUT
%token T_OUTPUT
%token T_DEREF
%token T_TRUE
%token T_FALSE
%token T_REPEAT
%token T_UNTIL

%token T_PARENTESIS_I
%token T_PARENTESIS_D
%token T_PUNTOCOMA
%token T_COMA
%token T_OPERADOR_SUMA
%token T_OPERADOR_RESTA
%token T_OPERADOR_MULT
%token T_OPERADOR_DIV
%token T_OPERADOR_IGUAL
%token T_OPERADOR_DESIGUAL
%token T_OPERADOR_MAYOR
%token T_OPERADOR_MAYOR_IGUAL
%token T_OPERADOR_MENOR
%token T_OPERADOR_MENOR_IGUAL
%token T_OPERADOR_ASIGNACION

%left T_OPERADOR_SUMA T_OPERADOR_RESTA
%left T_OPERADOR_MULT T_OPERADOR_DIV
%right NEG

%%

program:		T_BEGIN
			datos dcl_train
			codigo subroutines
			stm_train fin
			T_END
;

dcl_train:		declaration
		|	declaration dcl_train
;

declaration:		mode_def idlist T_PUNTOCOMA	{ }
;

mode_def:		mode				{ last_mode = $1; }
;

mode:			T_BOOL				{
		$$.clase = variable;
		$$.tipo = t_bool;
		$$.num_ref = 0;
	}
		|	T_INT				{
		$$.clase = variable;
		$$.tipo = t_int;
		$$.num_ref = 0;
	}
		|	T_REF mode			{
		$$ = $2;
		$$.num_ref ++;
	}
;

idlist:			id_def				{ }
		|	id_def T_COMA idlist		{ }
;

id_def:			T_ID				{
		if (nivel && last_mode.num_ref)
			error_semantico("puedes definir punteros "
					"solamente en el main");
		if (nivel)
			last_mode.num_param = num_param ++;

		if (!tabla_insertar(&tabla, $1.clave, copia(last_mode)))
			error_semantico("identificador redefinido");

		last_mode.clave = $1.clave;

		if (!nivel)
			asm_variable(last_mode);
	}
;

subroutines:		subroutine subroutines		{ }
		|	/* Vacio */			{ asm_main(); }
;

subroutine:		T_FUNCTION mode_def fn_def T_PARENTESIS_I
			fn_prm_train T_PARENTESIS_D fn_action T_BEGIN
			fn_dcl_train stm_train T_END	{
		asm_funcion_end();

		if (!retorno.num_ref)
			error_semantico("falta `return' en la funcion");

		tabla_cerrar(&tabla);
		nivel --;

		if (!tabla_insertar(&tabla, retorno.clave, copia(retorno)))
			error_semantico("nombre de funcion redefinido");

		free(retorno.clave);
	}
;

fn_def:			T_ID				{
		if (last_mode.num_ref)
			error_semantico(
				"el tipo de retorno no puede ser puntero");

		retorno = last_mode;
		retorno.clase = funcion;
		retorno.num_ref = 0;

		retorno.clave = malloc(strlen($1.clave) + 2);
		if (!retorno.clave) {
			error_semantico("malloc() fallo");
			retorno.clave = $1.clave;
		}

		/*
		 * Para el ambito usamos un nombre que no puede ser
		 * identificador en asple.
		 */
		sprintf(retorno.clave, ".%s", $1.clave);
		if (!tabla_abrir(&tabla, retorno.clave))
			error_semantico("funcion redefinida");
		strcpy(retorno.clave, $1.clave);

		retorno.parametros = 0;
		num_param = 0;
		nivel ++;
	}
;

fn_action:		/* Vacio */			{
		retorno.num_param = num_param;
		if (!tabla_insertar(&tabla, retorno.clave, copia(retorno)))
			error_semantico("identificador redefinido");

		num_param = 0;
	}
;

fn_prm_train:		fn_parameter rest_fn_prm_train	{ }
		|	/* Vacio */			{ }
;

rest_fn_prm_train:	T_PUNTOCOMA fn_parameter
			rest_fn_prm_train		{ }
		|	/* Vacio */			{ }
;

fn_parameter:		mode T_ID			{
		$1.clase = parametro;
		$1.num_param = num_param ++;

		if (!tabla_insertar(&tabla, $2.clave, copia($1)))
			error_semantico("identificador redefinido");

		nuevo_parametro($1);
	}
;

fn_dcl_train:		dcl_train			{
		asm_funcion(retorno, num_param);
	}
		|	/* Vacio */			{
		asm_funcion(retorno, num_param);
	}
;

stm_train:		statement			{ }
		|	statement T_PUNTOCOMA stm_train	{ }
;

statement:		asgt_stm			{ }
		|	cond_stm			{ }
		|	loop_stm			{ }
		|	transput_stm			{ }
		|	T_RETURN exp 			{
		if (!nivel)
			error_semantico("return fuera de funcion");
		else {
			retorno.num_ref ++;

			if (retorno.tipo != $2.tipo || $2.num_ref)
				error_semantico("tipo de retorno incorrecto");
		}

		asm_return();
	}
;

id_use:			T_ID				{
		if (tabla_buscar(&tabla, $1.clave, 0))
			$$ = * (token_t *) tabla_buscar(
					&tabla, $1.clave, &ambito);
		else {
			error_semantico("simbolo no definido");
			$$.tipo = t_int;
			$$.clase = variable;
			$$.clave = "unknown";
			$$.num_ref = 0;
		}

		if ($$.clase != funcion) {
			if (nivel && ambito && strcmp(ambito, "main")) {
				if ($$.clase == parametro)
					$$.num_param = retorno.num_param -
							$$.num_param;
				asm_variable_funcion($$);
			} else
				asm_variable_push($1.clave);
		} else {
			$$.clave = malloc(strlen($1.clave) + 1);
			$$.num_ref = 0;
			strcpy($$.clave, $1.clave);
		}

		last_mode = $$;
	}
;

asgt_stm:		id_use T_OPERADOR_ASIGNACION
			exp				{
		if ($1.clase != variable)
			error_semantico("asignacion de una no variable");
		if ($1.tipo != $3.tipo || $1.num_ref != $3.num_ref)
			error_semantico("tipos no compatibles");

		asm_asignacion();
	}
		|	dereference
			T_OPERADOR_ASIGNACION exp	{
		if ($1.tipo != $3.tipo || $1.num_ref != $3.num_ref)
			error_semantico("tipos no compatibles");

		asm_asignacion();
	}
;

cond_stm:		T_IF exp T_THEN if_stm_train
			T_FI				{
		if ($2.num_ref || $2.tipo != t_bool)
			error_semantico("se necesita expresion booleana");

		asm_fi($4.num_param);
	}
		|	T_IF exp T_THEN if_else_stm_train
			stm_train T_FI			{
		if ($2.num_ref || $2.tipo != t_bool)
			error_semantico("se necesita expresion booleana");

		asm_else_fi($4.num_param);
	}
;

if_else_stm_train:	if_stm_train T_ELSE		{
		asm_else($1.num_param);

		$$.num_param = $1.num_param;
	}
;

if_stm_train:		if_action stm_train		{
		$$.num_param = $1.num_param;
	}
;

if_action:		/* Vacio */			{
	 	asm_if(etiqueta);

		$$.num_param = etiqueta ++;
	}
;

loop_stm:		T_REPEAT repeat_action stm_train
			T_UNTIL exp			{
		if ($5.num_ref || $5.tipo != t_bool)
			error_semantico("until necesita expresion booleana");

		asm_if($2.num_param);
	}
		|	T_WHILE while_exp T_DO
			stm_train T_END			{
		asm_else($2.num_param);
	}
;

repeat_action:		/* Vacio */			{
		asm_fi(etiqueta);

		$$.num_param = etiqueta ++;
	}
;

while_exp:		while_action exp		{
		if ($2.num_ref || $2.tipo != t_bool)
			error_semantico("se necesita expresion booleana");

		asm_if($1.num_param);

		$$.num_param = $1.num_param;
	}
;

while_action:		/* Vacio */			{
		asm_else_fi(etiqueta);

		$$.num_param = etiqueta ++;
	}
;

transput_stm:		T_INPUT id_use			{
		if ($2.num_ref)
			error_semantico("asignacion de un puntero");
		if ($2.clase != variable)
			error_semantico("asignacion de una no variable");

		asm_input($2);
	}
		|	T_OUTPUT exp			{
		if ($2.num_ref)
			error_semantico("no se puede mostrar un puntero");

		asm_output($2);
	}
;

dereference:		T_DEREF id_use			{
		$$ = $2;
		if (/*$2.clase != variable || */!$2.num_ref)
			error_semantico(
				"esta expresion no puede ser dereferenciada");
		else
			$$.num_ref --;

		asm_desreferenciacion();
	}
		|	T_DEREF dereference 		{
		$$ = $2;
		if (!$2.num_ref)
			error_semantico(
				"esta expresion no puede ser dereferenciada");
		else
			$$.num_ref --;

		asm_desreferenciacion();
	}
;

exp:			exp T_OPERADOR_SUMA exp		{
		if ($1.tipo != $3.tipo || $1.num_ref + $3.num_ref)
			error_semantico("tipos no compatibles");
		$$ = $1;

		asm_suma($1);
	}
		|	exp T_OPERADOR_RESTA exp	{ 
		if ($1.tipo != $3.tipo || $1.num_ref + $3.num_ref)
			error_semantico("tipos no compatibles");
		$$ = $1;

		asm_resta($1);
	}
		|	exp T_OPERADOR_MULT exp		{ 
		if ($1.tipo != $3.tipo || $1.num_ref + $3.num_ref)
			error_semantico("tipos no compatibles");
		$$ = $1;

		asm_multiplicacion($1);
	}
		|	exp T_OPERADOR_DIV exp		{ 
		if ($1.tipo != $3.tipo || $1.num_ref + $3.num_ref)
			error_semantico("tipos no compatibles");
		$$ = $1;

		asm_division($1);
	}
		|	T_OPERADOR_RESTA exp %prec NEG	{
		if ($2.num_ref)
			error_semantico("negacion de un puntero");
		$$ = $2;

		asm_negacion($2);
	}
		|	id_use				{
		$$ = $1;

		/*
		 * En la pila esta ahora la direccion de la variable,
		 * lo unico que tenemos que hacer es una desreferenciacion
		 * porque la expresion es siempre una constante.
		 */
		asm_desreferenciacion();
	}
		|	constant			{
		$$ = $1;

		/* Las constantes tienen solamente valor */
		asm_valor_push($1.valor);
	}
		|	T_PARENTESIS_I exp
			T_PARENTESIS_D			{
		$$ = $2;
		/* No se genera ningun codigo */
	}
		|	T_PARENTESIS_I compare
			T_PARENTESIS_D			{
		$$ = $2;
		/* El valor ya esta en la pila */
	}
		|	id_use T_PARENTESIS_I explist
			T_PARENTESIS_D			{
		if ($1.clase != funcion)
			error_semantico("llamada a una no funcion");
		if ($1.num_param > $3.num_param)
			error_semantico("numero de parametros no suficiente");
		if ($1.num_param < $3.num_param)
			error_semantico("demasiados parametros");

		if ($1.clase == funcion)
			asm_llamada($1);
		free($1.clave);
	}
		|	dereference			{ 
		$$ = $1;

		asm_desreferenciacion();
	}
;

/*
 * En las cuatro siguientes producciones no hace falta
 * poner nada en la pila porque el valor ya esta ahi.
 *
 * Cambiamos el orden para evitar el uso de variables
 * globales. Con esto tambien evitamos la posiblidad de
 * que se desborde la pila del analizador porque con
 * recursividad de izquierda la pila no crece tanto.
 */
explist:		rest_explist			{
		$$ = $1;
	}
		|	/* Vacio */			{
		$$.num_param = 0;
	}
;

rest_explist:		rest_explist T_COMA exp		{
		$$ = $1;
		$$.num_param ++;

		if ($1.parametros) {
			if ($3.tipo != $1.parametros->tipo)
				error_semantico(
					"tipo de parametro no compatible");
			if ($3.num_ref != $1.parametros->num_ref)
				error_semantico(
					"tipo de parametro no compatible");

			$$.parametros = $1.parametros->parametros;
		} else
			$$.parametros = 0;
	}
		|	lpop exp			{
		$$.num_param = 1;

		if ($1.parametros) {
			if ($2.tipo != $1.parametros->tipo)
				error_semantico(
					"tipo de parametro no compatible");
			if ($2.num_ref != $1.parametros->num_ref)
				error_semantico(
					"tipo de parametro no compatible");

			$$.parametros = $1.parametros->parametros;
		} else
			$$.parametros = 0;
	}
;

lpop:			/* Vacio */			{
		$$ = last_mode;
	}
;

compare:		exp T_OPERADOR_IGUAL exp	{
		if ($1.tipo != t_int || $3.tipo != t_int ||
				$1.num_ref + $3.num_ref)
			error_semantico("tipos no compatibles en comparacion");
		$$.num_ref = 0;
		$$.tipo = t_bool;

		asm_igualdad();
	}
		|	exp T_OPERADOR_DESIGUAL exp	{
		if ($1.tipo != t_int || $3.tipo != t_int ||
				$1.num_ref + $3.num_ref)
			error_semantico("tipos no compatibles en comparacion");
		$$.num_ref = 0;
		$$.tipo = t_bool;

		asm_desigualdad();
	}
		|	exp T_OPERADOR_MENOR exp	{
		if ($1.tipo != t_int || $3.tipo != t_int ||
				$1.num_ref + $3.num_ref)
			error_semantico("tipos no compatibles en comparacion");
		$$.num_ref = 0;
		$$.tipo = t_bool;

		asm_menor();
	}
		|	exp T_OPERADOR_MENOR_IGUAL exp	{
		if ($1.tipo != t_int || $3.tipo != t_int ||
				$1.num_ref + $3.num_ref)
			error_semantico("tipos no compatibles en comparacion");
		$$.num_ref = 0;
		$$.tipo = t_bool;

		asm_menor_igual();
	}
		|	exp T_OPERADOR_MAYOR exp	{
		if ($1.tipo != t_int || $3.tipo != t_int ||
				$1.num_ref + $3.num_ref)
			error_semantico("tipos no compatibles en comparacion");
		$$.num_ref = 0;
		$$.tipo = t_bool;

		asm_mayor();
	}
		|	exp T_OPERADOR_MAYOR_IGUAL exp	{
		if ($1.tipo != t_int || $3.tipo != t_int ||
				$1.num_ref + $3.num_ref)
			error_semantico("tipos no compatibles en comparacion");
		$$.num_ref = 0;
		$$.tipo = t_bool;

		asm_mayor_igual();
	}
;

constant:		bool_constant			{
		$$ = $1;
	}
		|	int_constant			{
		$$ = $1;
	}
;

bool_constant:		T_TRUE				{
		$$.tipo = t_bool;
		$$.num_ref = 0;
		$$.valor = 1;
	}
		|	T_FALSE				{
		$$.tipo = t_bool;
		$$.num_ref = 0;
		$$.valor = 0;
	}
;

int_constant:		T_NUMBER			{
		$$.tipo = t_int;
		$$.num_ref = 0;
		$$.valor = $1.valor;
	}
;

datos:			/* Vacio */			{
		asm_segmento_datos();
	}
;

codigo:			/* Vacio */			{
		asm_segmento_codigo();
	}
;

fin:			/* Vacio */			{
		asm_segmento_fin();
	}
;

%%

token_t *copia(token_t t) {
	token_t *ret = malloc(sizeof(token_t));
	if (!ret) {
		error_semantico("malloc() fallo");
		exit(-1);
	}

	*ret = t;
	return ret;
}

int nuevo_parametro(token_t t) {
	token_t **i, *param = malloc(sizeof(token_t));
	if (!param) {
		error_semantico("malloc() fallo");
		exit(-1);
	}
	for (i = &retorno.parametros; *i; i = &(*i)->parametros);

	*param = t;
	*i = param;
	param->parametros = 0;
	return 0;
}
