#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <libgen.h>

#include "parse.h"
#include "simbolos.h"
#include "asm.h"

#define MODE	00644

const char *input_fn;
int lexico = 0;
int errores = 0;

void error_lexico(const char *msg) {
	fprintf(
			stderr,
			"%s: %i: error lexico - lexema no reconocido \"%s\"\n",
			input_fn,
			linea,
			msg);
	lexico = 1;
	errores ++;
}

void error_sintactico(const char *msg) {
	if (!lexico) {
		fprintf(
				stderr,
				"%s: %i: error sintactico\n",
				input_fn,
				linea);
		errores ++;
	}
}

void error_semantico(const char *msg) {
	fprintf(
			stderr,
			"%s: %i: error semantico - %s\n",
			input_fn,
			linea,
			msg);
	errores ++;
}

int main(int argc, char *argv[]) {
	int ret, fd;
	char fichero[512];

	/* La entrada y la salida */
	out = stdout; in = stdin;

	if (argc > 1) {
		input_fn = argv[1];
		if (!(in = fopen(argv[1], "r")))
			return fprintf(stderr, "Error de entrada\n");
	} else
		input_fn = "<stdin>";

	/* Preparacion de tabla de simbolos */
	tabla_crear(&tabla);

	asm_inicio();

	/* Analisis */
	ret = parse();

	tabla_liberar(&tabla);

	if (argc > 1)
		fclose(in);

	if (errores)
		fprintf(stderr, "%i error(es)\n", errores);
	else {	/* No hay errores, escribe el codigo */
		fd = 1;
		if (argc > 2) {
			if ((fd = creat(argv[2], MODE))
					== -1)
				return fprintf(stderr, "Error de salida\n");
		} else if (argc > 1) {
			sprintf(fichero, "%s.asm", basename(argv[1]));
			if ((fd = creat(fichero, MODE))
					== -1)
				return fprintf(stderr, "Error de salida\n");
		}

		asm_guardar(fd);

		if (argc > 1)
			close(fd);
	}

	/* Liberar recursos */
	asm_fin();

	return ret;
}
