/*
 * Implementacion tabla hash.
 */
#include <stdlib.h>
#include <string.h>

#include "hash.h"

hash_posicion_t hash_funcion(void *clave, int tamano_clave, int intento) {
	unsigned int seed;
	int pos;
	hash_posicion_t ret = 0;
	for (pos = 0; pos < tamano_clave; pos ++) {
		seed = ret + ((char *) clave)[pos];
		ret = rand_r(&seed);
	}
	for (pos = 0; pos < intento; pos ++)
		ret = rand_r(&seed);
	return ret;
}

int hash_crear(hash_t *tabla, int tamano) {
	int i;

	tabla->tamano = tamano;
	tabla->funcion = hash_funcion;
	tabla->tabla = malloc(tamano * sizeof(hash_elemento_t));
	if (!tabla->tabla)
		return 0; /* Falto memoria */

	for (i = 0; i < tamano; i ++)
		tabla->tabla[i].clave = 0;
	return 1;
}

int hash_liberar(hash_t *tabla) {
	free(tabla->tabla);
	return 1;
}

int hash_liberar_claves(hash_t *tabla) {
	int i;

	for (i = 0; i < tabla->tamano; i ++)
		if (tabla->tabla[i].clave) {
			free(tabla->tabla[i].clave);
			tabla->tabla[i].clave = 0;
		}
	return 1;
}

int hash_insertar(hash_t *tabla, hash_elemento_t *elemento) {
	int i = 0;
	hash_posicion_t pos = tabla->funcion(
			elemento->clave,
			elemento->tamano_clave,
			i);
	hash_posicion_t pos0 = pos;

	while (tabla->tabla[pos % tabla->tamano].clave) {
		if (tabla->tabla[pos % tabla->tamano].tamano_clave
			== elemento->tamano_clave && !memcmp(
			tabla->tabla[pos % tabla->tamano].clave,
			elemento->clave,
			elemento->tamano_clave))
			return 0; /* Clave ya existe */
		pos = tabla->funcion(
			elemento->clave,
			elemento->tamano_clave,
			++ i);
		if (pos == pos0)
			return 0; /* No hay espacio */
	}

	tabla->tabla[pos % tabla->tamano] = *elemento;
	return 1;
}

hash_elemento_t *hash_buscar(hash_t *tabla, void *clave, int tamano_clave) {
	int i = 0;
	hash_posicion_t pos = tabla->funcion(
			clave,
			tamano_clave,
			i);
	hash_posicion_t pos0 = pos;

	while (tabla->tabla[pos % tabla->tamano].clave && (
			tabla->tabla[pos % tabla->tamano].tamano_clave
			!= tamano_clave || memcmp(
			tabla->tabla[pos % tabla->tamano].clave,
			clave,
			tamano_clave))) { /* Mientras claves no coinciden */
		pos = tabla->funcion(
			clave,
			tamano_clave,
			++ i);
		if (pos == pos0)
			return 0; /* No existe elemento */
	}

	if (tabla->tabla[pos % tabla->tamano].clave)
		return &tabla->tabla[pos % tabla->tamano];

	return 0;
}
