/*
 * Implementacion de la tabla de simbolos.
 */
#include <stdlib.h>
#include <string.h>

#include "tabla.h"

int tabla_crear(tabla_simbolos_t *tabla) {
	if (!pila_crear(&tabla->ambitos))
		return 0;
	if (!hash_crear(&tabla->principal, TAMANO_TABLA))
		return 0;
	if (!pila_push(&tabla->ambitos, &tabla->principal, "main"))
		return 0;
	return 1;
}

int tabla_liberar(tabla_simbolos_t *tabla) {
	if (!pila_liberar(&tabla->ambitos))
		return 0;
	if (!hash_liberar_claves(&tabla->principal))
		return 0;
	if (!hash_liberar(&tabla->principal))
		return 0;
	return 1;
}

int tabla_insertar(tabla_simbolos_t *tabla, char *elemento, void *valor) {
	hash_elemento_t elem;
	elem.valor = valor;
	elem.clave = malloc(strlen(elemento));
	elem.tamano_clave = strlen(elemento);
	if (!elem.clave)
		return 0;
	memcpy(elem.clave, elemento, elem.tamano_clave);
	return hash_insertar(pila_top(&tabla->ambitos), &elem);
}

void *tabla_buscar(tabla_simbolos_t *tabla, char *elemento, char **amb) {
	hash_elemento_t *elem;
	pila_elemento_t *ambito;

	ambito = tabla->ambitos.cabeza;
	while (ambito) {
		elem = hash_buscar(
			ambito->valor0,
			elemento,
			strlen(elemento));
		if (amb)
			*amb = ambito->valor1;
		if (elem)
			return elem->valor;
		ambito = ambito->siguiente;
	}

	return 0;
}

int tabla_abrir(tabla_simbolos_t *tabla, char *ambito) {
	hash_t *n_ambito;
	hash_elemento_t elem;

	n_ambito = malloc(sizeof(hash_t));
	if (!n_ambito)
		return 0;
	if (!hash_crear(n_ambito, TAMANO_TABLA))
		return 0;

	elem.valor = n_ambito;
	elem.clave = malloc(strlen(ambito));
	elem.tamano_clave = strlen(ambito);
	if (!elem.clave)
		return 0;
	memcpy(elem.clave, ambito, elem.tamano_clave);
	if (!hash_insertar(pila_top(&tabla->ambitos), &elem))
		return 0;
	if (!pila_push(&tabla->ambitos, n_ambito, elem.clave))
		return 0;
	return 1;
}

int tabla_cerrar(tabla_simbolos_t *tabla) {
	return pila_pop(&tabla->ambitos);
}
